
'''
                   *********************************
                   *                               *
                   *     Intégration numérique     *
                   *                               *
                   *********************************
                   

'''

#Import des fonctions qui serviront

from math import pi 
import matplotlib.pyplot as plt
from random import random
import numpy as np


'''
1) Méthode des rectangles à gauche
'''

def MethRectG(f,a,b,n):
    # f est une fonction définie sur l'intervalle [a;b] avec a<b
    # n est un entier naturel non nul
    # la fonction renvoie l'approximation de l'intégrale de f entre a et b
    # en utilisant la méthode des rectangles à gauche avec n rectangles 
    
    pas=(b-a)/n   # on crée le pas de la subdivision = taille d'un petit
                    # intervalle = longueur d'un rectangle (axe abscisses)
    x=a   # subdivision, initialisée à a 
    somme=0   # résultat en construction, on l'intialise à 0
    

    for i in range(n): # on parcourt les n petits rectangles
        # range(n) = intervalle entier [0,n-1] = [0,1,2 ... , n-1]
        
        somme = somme + pas*f(x) # on rajoute l'aire du nouveau
                                # petit rectangle = longueur sur l'axe
                                # des abscisses * hauteur
        
        x=x+pas  # on passe au terme suivant de la subdivision 
        
        # on a fini la boucle for
        
    return(somme) # la somme contient l'approximation de l'intégrale
                    # de f entre a et b par n rectangles

def test_fonction(x):
    return(x)
    

valeurExacte=pi/20 


def fonctionRunge(x):
    # x est un réel, renvoie l'image de x par la fonction de Runge
    return(1/(25+x**2))


def erreurRectG():
    # provoque l'affichage de l'erreur commise par la méthode des rectangles à gauche
    # en fonction du nombre de rectangles
    
    N=[10,20,30,50,100,150,300,500,1000,1500,2000,5000,10000]   # nombres de rectangles
    
    erreurs=[]  # intialisation des erreurs
    
    for n in N :
        
        erreurs.append(abs(MethRectG(fonctionRunge,0,5,n)-valeurExacte))
        # la valeur absolue pour mesurer la distance entre les nombres
        
    plt.plot(N,erreurs,'-b',label='rectangles à gauche')
    plt.legend( loc='upper right', numpoints = 1 )
    plt.xlabel('nombre de points')
    plt.ylabel('\'erreur')
    plt.show()



def erreurRectGLog():
    # provoque l'affichage de l'erreur commise par la méthode des rectangles à gauche
    # en fonction du nombre de rectangles
    
    N=[10,20,30,50,100,150,300,500,1000,1500,2000,5000,10000]   # nombres de rectangles
    
    erreurs=[]  # intialisation des erreurs
    
    for n in N :
        
        erreurs.append(abs(MethRectG(fonctionRunge,0,5,n)-valeurExacte))
        # la valeur absolue pour mesurer la distance entre les nombres
        
        ### Attention np.log = logarithme népérien
        
        
    plt.plot(np.log(N),np.log(erreurs),'-b',label='rectangles à gauche')
    plt.legend( loc='upper right', numpoints = 1 )
    plt.xlabel('log(nombre de points)')
    plt.ylabel('\'log(erreur)')
    plt.show()
    
    

'''
2) Autres méthodes d'intégration numérique et comparaison
'''



def MethRectD(f,a,b,n):
    # f est une fonction définie sur l'intervalle [a;b] avec a<b
    # n est un entier naturel non nul
    # la fonction renvoie l'approximation de l'intégrale de f entre a et b
    # en utilisant la méthode des rectangles à gauche avec n rectangles 
    
    pas=(b-a)/n   # on crée le pas de la subdivision
    x=a   # subdivision, initialisée à a 
    somme=0   # résultat en construction, on l'intialise à 0
    

    for i in range(n):
        x=x+pas  
        somme = somme + pas*f(x) 
    return(somme)
    


def methPtMilieu(f,a,b,n):
    # f est une fonction, a<b des réels, n>0 un entier.
    # renvoit l'approximation de l'intégrale de f entre a et b obtenue 
    # avec la méthode du point milieu
    pas=(b-a)/n
    integrale=0
    for i in range(n):
        integrale=integrale+pas*f(a+i*pas+pas/2)
    return(integrale)
    
    
def methTrapeze(f,a,b,n):
    # f est une fonction, a<b des réels, n>0 un entier.
    # renvoit l'approximation de l'intégrale de f entre a et b obtenue 
    # avec la méthode des trapezes (naïf : pas optimisé sur les calculs d'images)
    pas=(b-a)/n
    integrale=0
    for i in range(n):
        integrale=integrale+pas*(f(a+i*pas)+f(a+(i+1)*pas))*.5
    return(integrale)

    
def compareErreur():
    # provoque l'affichage d'un graphique des courbes des erreurs des différentes
    # méthodes d'intégration en fonction du nombre de points utilisés.
    
    # nombres d'intervalles élémentaires des subdivisions : 
    N=[10,20,30,50,100,150,300,500,1000,1500,2000,5000,10000]
    
    # initialisation des erreurs :
    erreur_RectG=[]
    erreur_RectD=[]
    erreur_pointM=[]
    erreur_Trapeze=[]
    
    for n in N :
        erreur_RectG.append(abs(valeurExacte-MethRectG(fonctionRunge,0,5,n)))
        erreur_RectD.append(abs(valeurExacte-MethRectD(fonctionRunge,0,5,n)))
        erreur_pointM.append(abs(valeurExacte-methPtMilieu(fonctionRunge,0,5,n)))
        erreur_Trapeze.append(abs(valeurExacte-methTrapeze(fonctionRunge,0,5,n)))
        
    plt.plot(np.log(N),np.log(erreur_RectG),'-r',label='rectangles à gauche')
    plt.plot(np.log(N),np.log(erreur_RectD),'-b',label='rectangles à droite')
    plt.plot(np.log(N),np.log(erreur_pointM),'-c',label='point milieu')
    plt.plot(np.log(N),np.log(erreur_Trapeze),'-g',label='trapèzes')
    
    plt.legend( loc='lower left', numpoints = 1 )
    plt.xlabel('logarithme du nombre de points')
    plt.ylabel('logarithme de l\'erreur')
    plt.show()
    
    
# pour les rectangles, l'erreur est en O(1/n) ce qui donne ln(erreur)=a-ln(n) : on a donc les points sur une droite de pente -1
# pour le point milieu et les trapezes, l'erreur est en O(1/n^2) ce qui donne ln(erreur)=a-2ln(n) : on a donc les points sur une droite de pente -2
    


# Mise en œuvre avec des fonctions élémentaires et une fonction qui les compose pour avoir les 
# méthodes globales.
    

def integ_elem_rectG(f,a,b):
    return(f(a)*(b-a))

def integ_elem_mil(f,a,b):
    return(f((a+b)/2)*(b-a))

def integ_elem_trapeze(f,a,b):
    return((f(a)+f(b))/2*(b-a))
    
def integ_comp(methelementaire,f,a,b,n):
    pas=(b-a)/n
    approx=0
    for i in range(n):
        approx=approx+methelementaire(f,a+i*pas,a+(i+1)*pas)
    return(approx)
        




'''
3) Méthode de Simpson
'''


# On programme la méthode élémentaire pour Simpson, la méthode globale utilise la fonction de composition

def integ_elem_simpson(f,a,b):
    return((f(a)+f(b)+4*f((a+b)/2))/6*(b-a))    
    


def compareSimpson():
    # provoque l'affichage d'un graphique des courbes des erreurs des différentes
    # méthodes d'intégration en fonction du nombre de points utilisés.
    N=[10,20,30,50,100,150,300,500,1000]
    erreur_RectG=[]
    erreur_pointM=[]
    erreur_Trapeze=[]
    erreur_Simpson=[]
    
    for n in N :
        erreur_RectG.append(abs(valeurExacte-MethRectG(fonctionRunge,0,5,n)))
        erreur_pointM.append(abs(valeurExacte-methPtMilieu(fonctionRunge,0,5,n)))
        erreur_Trapeze.append(abs(valeurExacte-methTrapeze(fonctionRunge,0,5,n)))
        erreur_Simpson.append(abs(valeurExacte-integ_comp(integ_elem_simpson,fonctionRunge,0,5,n)))
    
    plt.plot(np.log(N),np.log(erreur_RectG),'-b',label='rectangles à gauche')
    plt.plot(np.log(N),np.log(erreur_pointM),'-c',label='point milieu')
    plt.plot(np.log(N),np.log(erreur_Trapeze),'-g',label='trapèzes')
    plt.plot(np.log(N),np.log(erreur_Simpson),'-r',label='Simpson')
    
    plt.legend( loc='lower right', numpoints = 1 )
    plt.xlabel('logarithme du nombre de points')
    plt.ylabel('logarithme de l\'erreur')
    plt.show()
    
    
# Lorsque l'erreur devient de l'ordre de 10^(-16), on atteint les limites du type float. Autrement dit  :
# le type float ne permet plus de faire la différence entre les nombres. Ceci explique l'allure de la courbe.
    


'''
4) Méthodes de Monte Carlo
'''


# POUR LES INTEGRALES

def methMonteCarlo(f,a,b,n):
    # f une fonction intégrable sur [a;b] , n>0 un entier naturel
    # renvoie une approximation de l'intégrale de f sur [a;b] obtenue par
    # la méthode de Monte-Carlo
    valMoy=0
    for i in range(n):
        alea=a+(b-a)*random()
        valMoy=valMoy+f(alea)
    valMoy=valMoy/n
    return(valMoy*(b-a))


def compareMonteCarlo():
    # provoque l'affichage d'un graphique des courbes des erreurs des différentes
    # méthodes d'intégration en fonction du nombre de points utilisés.
    
    # nombres d'intervalles élémentaires des subdivisions : 
    N=[10,20,30,50,100,150,300,500,1000,1500,2000,5000,10000]
    
    # initialisation des erreurs :
    erreur_RectG=[]
    erreur_pointM=[]
    erreur_Trapeze=[]
    erreur_MC=[]

    
    for n in N :
        erreur_RectG.append(abs(valeurExacte-MethRectG(fonctionRunge,0,5,n)))
        erreur_pointM.append(abs(valeurExacte-methPtMilieu(fonctionRunge,0,5,n)))
        erreur_Trapeze.append(abs(valeurExacte-methTrapeze(fonctionRunge,0,5,n)))
        erreur_MC.append(abs(valeurExacte-methMonteCarlo(fonctionRunge,0,5,n)))

    plt.plot(np.log(N),np.log(erreur_RectG),'-r',label='rectangles à gauche')
    plt.plot(np.log(N),np.log(erreur_pointM),'-c',label='point milieu')
    plt.plot(np.log(N),np.log(erreur_Trapeze),'-g',label='trapèzes')
    plt.plot(np.log(N),np.log(erreur_MC),'-b',label='MonteCarlo')

    
    plt.legend( loc='lower left', numpoints = 1 )
    plt.xlabel('logarithme du nombre de points')
    plt.ylabel('logarithme de l\'erreur')
    plt.show()
    
    
    # On observe que la méthode de Monte Carlo n'est pas plus performante que celle
    # des rectangles.
    

# SURFACE DE L'ELLIPSE
    
    
def estDansS(x,y):
    # (x,y) sont les coordonnées d'un point du plan
    # la fonction renvoie True si le point considéré est dans la surface 
    # S:x^2+3y^2<=1, False sinon.
    return(x**2+3*y**2<=1)
    
def approxS(n):
    # n est un int
    # renvoie l'approximation obtenue avec la méthode de Monte Carlo pour S
    nombre_dans_S=0
    for i in range(n):
        x=2*random()-1
        y=2*random()-1
        if estDansS(x,y):
            nombre_dans_S=nombre_dans_S+1
            
    proportion=nombre_dans_S/n
    
    return(proportion*4)    
    
# on observe une assez forte instabilité du résultat : avec 10^4 points, on a seulement 2 décimales
# qui sont "stables". On peut un peu améliorer les choses en prenant un rectangle plus petit qui contient
# S (sur l'axe des ordonnées, on peut varier uniquement entre +/- 1/sqrt(3))
   
   
def approxS_bis(n):
    # n est un int
    # renvoie l'approximation obtenue avec la méthode de Monte Carlo pour S
    nombre_dans_S=0
    inv_racine3=3**(-.5)
    for i in range(n):
        x=2*random()-1
        y=2*inv_racine3*random()-inv_racine3
        if estDansS(x,y):
            nombre_dans_S=nombre_dans_S+1
            
    proportion=nombre_dans_S/n
    
    return(proportion*4*inv_racine3)    
    
